class Bindings(object):
    def __init__(self):
        self._bindings = {}
    
    def bind(self, key, provider=None):
        if key in self:
            raise AlreadyBoundException("Cannot rebind key: %s" % key)
            
        if provider is None:
            return Binder(key, self)
        else:
            self._bindings[key] = provider
    
    def copy(self):
        copy = Bindings()
        copy._bindings = self._bindings.copy()
        return copy
    
    def update(self, bindings):
        for key in bindings._bindings:
            if key in self._bindings:
                raise AlreadyBoundException("Key already bound: %s" % key)
        self._bindings.update(bindings._bindings)
    
    def __contains__(self, key):
        return key in self._bindings
        
    def __getitem__(self, key):
        return self._bindings[key]

class Binder(object):
    def __init__(self, key, bindings):
        self._key = key
        self._bindings = bindings
    
    def to_instance(self, instance):
        self.to_provider(lambda injector: instance)
    
    def to_key(self, key):
        if key is self._key:
            raise TypeError("Cannot bind a key to itself")
        self.to_provider(lambda injector: injector.get(key))
    
    def to_type(self, key):
        return self.to_key(key)
    
    def to_provider(self, provider):
        self._bindings.bind(self._key, provider)

class AlreadyBoundException(Exception):
    pass



from nose.tools import assert_equals
from nose.tools import assert_raises

from zuice import Bindings
from zuice import Injector
from zuice import NoSuchBindingException
from zuice import dependency
from zuice import Base

class Apple(object):
    pass
    
default_apple = Apple()
    
class Banana(object):
    pass
    
def test_bind_type_to_instance():
    apple = Apple()
    bindings = Bindings()
    bindings.bind(Apple).to_instance(apple)
    
    injector = Injector(bindings)
    assert injector.get(Apple) is apple

def test_bind_name_to_instance():
    apple = Apple()
    bindings = Bindings()
    bindings.bind("apple").to_instance(apple)
    
    injector = Injector(bindings)
    assert injector.get("apple") is apple
    
def test_bind_type_to_provider():
    apple = Apple()
    bindings = Bindings()
    bindings.bind(Apple).to_provider(lambda injector: apple)
    
    injector = Injector(bindings)
    assert injector.get(Apple) is apple


def test_get_throws_exception_if_no_such_binding_exists_and_object_has_init_args():
    class Donkey(object):
        def __init__(self, legs):
            pass
        
    injector = Injector(Bindings())
    assert_raises(NoSuchBindingException, lambda: injector.get(Donkey))
    
def test_get_raises_exception_if_no_such_binding_exists():
    injector = Injector(Bindings())
    assert_raises(NoSuchBindingException, lambda: injector.get("apple"))
    
    apple = Apple()
    bindings = Bindings()
    bindings.bind("apple").to_provider(lambda injector: apple)
    
    injector = Injector(bindings)
    assert injector.get("apple") is apple
    assert_raises(NoSuchBindingException, lambda: injector.get("banana"))

def test_changing_bindings_after_creating_injector_does_not_change_injector():
    bindings = Bindings()
    injector = Injector(bindings)
    bindings.bind("apple").to_instance(Apple())
    assert_raises(NoSuchBindingException, lambda: injector.get("apple"))


def test_can_inject_class_with_no_constructor_arguments():
    class Coconut(object):
        def __init__(self):
            self.x = 10
            
    injector = Injector(Bindings())
    coconut = injector.get(Coconut)
    assert_equals(10, coconut.x)


def test_can_bind_to_names():
    apple_to_inject = Apple()
    bindings = Bindings()
    bindings.bind("apple").to_instance(apple_to_inject)
    bindings.bind("another_apple").to_key("apple")
    
    injector = Injector(bindings)
    assert injector.get("another_apple") is apple_to_inject


def test_injector_class_is_bound_to_injector():
    injector = Injector(Bindings())
    assert injector.get(Injector) is injector
    

def test_classes_that_inherit_from_injectable_have_members_injected():
    class Foo(Base):
        _tag_fetcher = dependency("tag_fetcher")
    
    tag_fetcher = {'some': 'object'}
    
    bindings = Bindings()
    bindings.bind("tag_fetcher").to_instance(tag_fetcher)
    injector = Injector(bindings)
    assert injector.get(Foo)._tag_fetcher is tag_fetcher

def test_classes_that_inherit_from_injectable_can_be_passed_constructor_arguments_manually_by_name():
    class Foo(Base):
        fetcher = dependency("tag_fetcher")
    
    tag_fetcher = {'some': 'object'}
    
    assert Foo(fetcher=tag_fetcher).fetcher is tag_fetcher

def test_injectable_members_have_leading_underscores_removed_in_constructor_arg():
    class Foo(Base):
        _fetcher = dependency("tag_fetcher")
    
    tag_fetcher = {'some': 'object'}
    
    assert Foo(fetcher=tag_fetcher)._fetcher is tag_fetcher

def test_classes_that_inherit_from_injectable_can_be_passed_constructor_arguments_manually_by_position():
    class View(Base):
        _tag_fetcher = dependency("tag_fetcher")
        _post_fetcher = dependency("post_fetcher")
    
    tag_fetcher = {'some': 'object'}
    post_fetcher = {'another': 'object'}
    
    view = View(tag_fetcher, post_fetcher)
    assert view._tag_fetcher is tag_fetcher
    assert view._post_fetcher is post_fetcher

def test_injecting_overspecified_arguments_to_injectable_raises_exception():
    class View(Base):
        _tag_fetcher = dependency("tag_fetcher")
    
    tag_fetcher = {'some': 'object'}
    
    try:
        view = View(tag_fetcher, tag_fetcher=tag_fetcher)
        assert False
    except TypeError, e:
        assert_equals(str(e), "Got multiple values for keyword argument 'tag_fetcher'")

def test_injecting_too_many_positional_arguments_to_injectable_raises_exception():
    class View(Base):
        _tag_fetcher = dependency("tag_fetcher")
    
    try:
        view = View(None, None)
        assert False
    except TypeError as e:
        assert_equals(str(e), "__init__ takes exactly 2 arguments (3 given)")

def test_injectable_injects_attributes_of_sub_classes():
    class Parent(Base):
        _tag_fetcher = dependency('tag_fetcher')
        
    class Child(Parent):
        _blog_post_fetcher = dependency('post_fetcher')

    tag_fetcher = {'some': 'object'}
    post_fetcher = {'another': 'object'}
    
    bindings = Bindings()
    bindings.bind("tag_fetcher").to_instance(tag_fetcher)
    bindings.bind("post_fetcher").to_instance(post_fetcher)
    injector = Injector(bindings)
    child = injector.get(Child)
    
    assert child._tag_fetcher is tag_fetcher
    assert child._blog_post_fetcher is post_fetcher

def test_subclassing_injectable_objects_allows_injected_attributes_to_be_overwritten():
    class Parent(Base):
        _fetcher = dependency('tag_fetcher')
        
    class Child(Parent):
        _fetcher = dependency('post_fetcher')

    post_fetcher = {'another': 'object'}
    
    bindings = Bindings()
    bindings.bind("post_fetcher").to_instance(post_fetcher)
    injector = Injector(bindings)
    child = injector.get(Child)
    
    assert child._fetcher is post_fetcher
    
def test_missing_constructor_arguments_in_injectable_raises_type_error():
    class Foo(Base):
        _tag_fetcher = dependency("tag_fetcher")
        _blog_post_fetcher = dependency('post_fetcher')
    
    tag_fetcher = {'some': 'object'}
    
    assert_raises(TypeError, lambda: Foo(_tag_fetcher=tag_fetcher))

def test_injectable_injecting_manually_with_extra_members_raises_type_error():
    class Foo(Base):
        _tag_fetcher = dependency("tag_fetcher")
    
    tag_fetcher = {'some': 'object'}
    post_fetcher = {'another': 'object'}
    
    assert_raises(TypeError, lambda: Foo(_tag_fetcher=tag_fetcher, _post_fetcher=post_fetcher))
    



import inspect


def has_no_arg_constructor(cls):
    constructor = cls.__init__
    
    if constructor is object.__init__:
        return True
    
    arg_specs = inspect.getargspec(constructor)
    arg_names = arg_specs[0]
    return len(arg_names) == 1



import itertools

import zuice.reflect
from .bindings import Bindings

__all__ = ['Bindings', 'Injector', 'Base', 'dependency']

class Injector(object):
    def __init__(self, bindings):
        self._bindings = bindings.copy()
        self._bindings.bind(Injector).to_instance(self)
    
    def get(self, key):
        if key in self._bindings:
            return self._bindings[key](self)
            
        elif isinstance(key, type):
            return self._get_from_type(key)
        
        else:
            raise NoSuchBindingException(key)
    
    def _get_from_type(self, type_to_get):
        if hasattr(type_to_get.__init__, 'zuice'):
            return self._inject(type_to_get, type_to_get.__init__.zuice)
        if zuice.reflect.has_no_arg_constructor(type_to_get):
            return type_to_get()
        raise NoSuchBindingException(type_to_get)
    
    def _inject(self, to_call, argument_builder):
        args, kwargs = argument_builder.build_args(self)
        return to_call(*args, **kwargs)
    
class NoSuchBindingException(Exception):
    def __init__(self, key):
        self.key = key
        
    def __str__(self):
        return str(self.key)


class Dependency(object):
    _counter = itertools.count()
    
    def __init__(self, key):
        self._key = key
        self._ordering = self._counter.next()
        
    def inject(self, injector):
        return injector.get(self._key)

def dependency(key):
    return Dependency(key)

class InjectableConstructor(object):
    def build_args(self, injector):
        return [], {"___injector": injector}

class Base(object):
    def __init__(self, *args, **kwargs):
        attrs = []
        for key in dir(type(self)):
            attr = getattr(self, key)
            if isinstance(attr, Dependency):
                attrs.append((key, attr))
            
        if '___injector' in kwargs:
            injector = kwargs.pop('___injector')
            for key, attr in attrs:
                setattr(self, key, attr.inject(injector))
        else:
            if len(args) > len(attrs):
                raise TypeError(
                    "__init__ takes exactly %s arguments (%s given)" %
                        (len(attrs) + 1, len(args) + 1)
                )
            attrs.sort(key=lambda (key, attr): attr._ordering)
            for index, (key, attr) in enumerate(attrs):
                arg_name = key.lstrip("_");
                
                if index < len(args):
                    if arg_name in kwargs:
                        raise TypeError("Got multiple values for keyword argument '%s'" % arg_name)
                    setattr(self, key, args[index])
                elif arg_name in kwargs:
                    setattr(self, key, kwargs.pop(arg_name))
                else:
                    raise TypeError("Missing keyword argument: %s" % key)
        
        if len(kwargs) > 0:
            raise TypeError("Unexpected keyword argument: " + kwargs.items()[0][0])
    
    __init__.zuice = InjectableConstructor()



import zuice.reflect

def test_class_with_no_explicit_constructor_has_no_arg_constructor():
    class Empty(object):
        pass
        
    assert zuice.reflect.has_no_arg_constructor(Empty)


def test_class_no_arg_constructor_is_recognised_as_no_arg_constructor():
    class Boring(object):
        def __init__(self):
            pass
        
    assert zuice.reflect.has_no_arg_constructor(Boring)



def test_class_constructor_with_args_is_not_recognised_as_no_arg_constructor():
    class SampleObject(object):
        def __init__(self, first, second=None, third=30):
            pass

    assert not zuice.reflect.has_no_arg_constructor(SampleObject)



from nose.tools import assert_raises
from nose.tools import assert_equals

from zuice.bindings import AlreadyBoundException
from zuice.bindings import Bindings

class Apple(object):
    pass

def test_key_in_bindings_if_key_has_been_bound():
    apple = Apple()
    bindings = Bindings()
    bindings.bind("apple").to_instance(apple)
    
    assert "apple" in bindings
    assert "banana" not in bindings

def test_cannot_bind_using_the_same_binder_more_than_once():
    apple = Apple()
    bindings = Bindings()
    binder = bindings.bind(Apple)
    binder.to_provider(lambda injector: apple)
    assert_raises(AlreadyBoundException, lambda: binder.to_provider(lambda injector: apple))

def test_cannot_bind_the_same_key_more_than_once():
    apple = Apple()
    bindings = Bindings()
    bindings.bind('apple').to_instance(apple)
    assert_raises(AlreadyBoundException, lambda: bindings.bind('apple'))
    assert_raises(AlreadyBoundException, lambda: bindings.bind('apple'))

def test_cannot_bind_key_to_itself():
    bindings = Bindings()
    assert_raises(TypeError, lambda: bindings.bind("apple").to_key("apple"))

def test_can_update_bindings_with_more_bindings():
    bindings = Bindings()
    bindings.bind("maximum_threads").to_instance(5)
    
    new_bindings = Bindings()
    new_bindings.bind("minimum_threads").to_instance(2)
    
    bindings.update(new_bindings)
    assert_equals(bindings["maximum_threads"](None), 5)
    assert_equals(bindings["minimum_threads"](None), 2)

def test_cannot_override_existing_bindings():
    bindings = Bindings()
    bindings.bind("maximum_threads").to_instance(5)
    
    new_bindings = Bindings()
    new_bindings.bind("maximum_threads").to_instance(2)
    
    assert_raises(AlreadyBoundException, lambda: bindings.update(new_bindings))



from distutils.core import setup

setup(
    name='zuice',
    version='0.2.0',
    description='A dependency injection framework for Python',
    author='Michael Williamson',
    author_email='mike@zwobble.org',
    url='https://github.com/mwilliamson/zuice',
    packages=['zuice'],
    keywords="dependency injection di",
)
